#open "misc" ;;
#open "treetype" ;;

(* USED WHEN NEEDED TO SEE IF ALREADY DEFINED *)
let produce_and_mark_ident_type env = function
   Idnt_simple (s1,s2) ->
       emit_string s2 ;
       begin
       match is_defined s1 Obj_type env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1
       end
 | Idnt_prefix (s1,s2,s3) ->
       emit_string (s2^"<B>prefix</B>") ;
       emit_string s3 ;
       begin
       match is_defined s1 Obj_type env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1
       end
;;

let produce_and_mark_ident_value env = function
   Idnt_simple (s1,s2) ->
       emit_string s2 ;
       begin
       match is_defined s1 Obj_value env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1
       end
 | Idnt_prefix (s1,s2,s3) ->
       emit_string (s2^"<B>prefix</B>") ;
       emit_string s3 ;
       begin
       match is_defined s1 Obj_value env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1
       end
;;

let produce_and_mark_extid_type env = function
   Eid_qual (s1,id,s2,s3) ->
      emit_string s2 ; emit_string s1;emit_string s3;emit_string "__" ;
      produce_and_mark_ident_type env id ;
 | Eid_simple id ->
      produce_and_mark_ident_type env id
;;

let produce_and_mark_extid_value env = function
   Eid_qual (s1,id,s2,s3) ->
      emit_string s2;emit_string s1;emit_string s3;emit_string "__";
      produce_and_mark_ident_value env id ;
 | Eid_simple id ->
      produce_and_mark_ident_value env id
;;


(* USED WHEN DEFINING A NEW ANCHORAGE *)
let produce_and_store_ident_type env = function
   Idnt_simple (s1,s2) ->
       let mark = newmarker () in
        emit_string s2 ;
        emit_aname (s1^(string_of_int mark)) s1 ;
        [(s1,Obj_type,mark)]
 | Idnt_prefix (s1,s2,s3) ->
       let mark = newmarker () in
        emit_string (s2^"<B>prefix</B>") ;
        emit_string s3 ;
        emit_aname (s1^(string_of_int mark)) s1 ;
        [(s1,Obj_type,mark)]
;;

let produce_and_store_ident_value env = function
   Idnt_simple (s1,s2) ->
       let mark = newmarker () in
        emit_string s2 ;
        emit_aname (s1^(string_of_int mark)) s1 ;
        [(s1,Obj_value,mark)]
 | Idnt_prefix (s1,s2,s3) ->
       let mark = newmarker () in
        emit_string (s2^"<B>prefix</B>") ;
        emit_string s3 ;
        emit_aname (s1^(string_of_int mark)) s1 ;
        [(s1,Obj_value,mark)]
;;

let produce_and_store_extid_type env = function
   Eid_qual (s1,id,s2,s3) ->
      emit_string s2 ; emit_string s1;emit_string s3;emit_string "__" ;
      produce_and_store_ident_type env id ;
 | Eid_simple id ->
      produce_and_store_ident_type env id
;;

let produce_and_store_extid_value env = function
   Eid_qual (s1,id,s2,s3) ->
      emit_string s2;emit_string s1;emit_string s3;emit_string "__";
      produce_and_store_ident_value env id ;
 | Eid_simple id ->
      produce_and_store_ident_value env id
;;

(* USED WHEN WE DON'T WANT ANYTHING ELSE BUT PRINTING THE TOKEN *)
let produce_ident env = function
   Idnt_simple (s1,s2) ->
       emit_string (s2^s1)
 | Idnt_prefix (s1,s2,s3) ->
       emit_string (s2^"<B>prefix</B>"^s3^s1)
;;

let produce_extid env = function
   Eid_qual (s1,id,s2,s3) ->
      emit_string s2;emit_string s1;emit_string s3;emit_string "__";
      produce_ident env id ;
 | Eid_simple id ->
      produce_ident env id
;;







let produce_atomcst env = function
   Acst_int (i,s) ->
       emit_string s ;
       emit_string ("<FONT COLOR=\"#"^(!intcolor)^"\">") ;
       emit_int i ;
       emit_string "</FONT>"
 | Acst_float (f,s) ->
       emit_string s ;
       emit_string ("<FONT COLOR=\"#"^(!floatcolor)^"\">") ;
       emit_float f ;
       emit_string "</FONT>"
 | Acst_string (s1,s2) ->
       emit_string ("<FONT COLOR=\"#"^(!stringcolor)^"\">") ;
       emit_string (s2^"\""^s1^"\"") ;
       emit_string "</FONT>"
 | Acst_char (c,s) ->
       emit_string s ;
       emit_string ("<FONT COLOR=\"#"^(!charcolor)^"\">") ;
       emit_string "`" ;
       emit_string c ;
       emit_string "`" ;
       emit_string "</FONT>"
;;

let rec produce_stype env = function
   Styp_var (s1,s2,s3) ->
       emit_string s2 ;
       emit_string "<B>'</B>" ;
       emit_string s3 ;
       emit_string s1
 | Styp_extid eid ->
       produce_and_mark_extid_type env eid
 | Styp_typapp (styp,eid) ->
       produce_stype env styp ;
       produce_and_mark_extid_type env eid
 | Styp_typappmulti (typ,typl,eid,s1,s2,s3) ->
       emit_string (s1^"<B>(</B>") ;
       produce_type env typ ;
       emit_string (s2^"<B>,</B>") ;
       do_seplist (produce_type env) typl ;
       emit_string (s3^"<B>)</B>") ;
       produce_and_mark_extid_type env eid
 | Styp_paren (typ,s1,s2) ->
       emit_string (s1^"<B>(</B>") ;
       produce_type env typ ;
       emit_string (s2^"<B>)</B>")

and produce_type env = function
   Typ_simple styp ->
      produce_stype env styp
 | Typ_starlst stypl ->
      do_seplist (produce_stype env) stypl
 | Typ_minusgreater (typ1,typ2,s) ->
      produce_type env typ1 ;
      emit_string (s^"<B>-&gt;</B>") ;
      produce_type env typ2

and produce_patlabel env = function
   Patl_underscore s ->
       emit_string (s^"_") ;
       []
 | Patl_other (eid,pat,s) ->
       produce_and_mark_extid_value env eid ;
       emit_string (s^"<B>=</B>") ;
       produce_pattern env pat

and produce_pattern env = function
   Pat_spat spat ->
      produce_spattern env spat
 | Pat_asid (pat,s1,s2,s3) ->
      let newdefs = produce_pattern env pat in
       emit_string (s2^"<B>as</B>") ;
       emit_string s3 ;
       let mark = newmarker () in
        emit_aname (s1^(string_of_int mark)) s1 ;
        [(s1,Obj_value,mark)]@newdefs
 | Pat_coloncolon (pat1,pat2,s) ->
      let newdefs1 = produce_pattern env pat1 in
       emit_string (s^"<B>::</B>") ;
       let newdefs2 = produce_pattern env pat2 in
        newdefs2@newdefs1
 | Pat_commalst patl ->
      emit_list (produce_pattern env) patl
 | Pat_extid (eid,spat) ->
      (* USED FOR CONSTRUCT VALUES *)
      produce_extid env eid ;
      let newdefs = produce_spattern env spat in
       newdefs
 | Pat_multi (pat1,pat2,s) ->
      let newdefs1 = produce_pattern env pat1 in
       emit_string (s^"<B>|</B>") ;
       let newdefs2 = produce_pattern env pat2 in
        newdefs2@newdefs1

and produce_spattern env = function
   Spat_atcst atom ->
       produce_atomcst env atom ;
       []
 | Spat_subsint (s1,i,s2,s3) ->
       emit_string (s2^s1) ;
       emit_string s3 ;
       emit_int i ;
       []
 | Spat_subfloat (s1,f,s2,s3) ->
       emit_string (s2^s1) ;
       emit_string s3 ;
       emit_float f ;
       []
 | Spat_underscore s ->
       emit_string (s^"_") ;
       []
 | Spat_ident id ->
      let newdefs = produce_and_store_ident_value env id in
       newdefs
 | Spat_qualid (s1,id,s2,s3) ->
       emit_string s2 ;
       emit_string s1 ;
       emit_string s3 ;
       emit_string "__" ;
       produce_and_mark_ident_value env id ;
       []
 | Spat_unit (s1,s2) ->
       emit_string (s1^"<B>(</B>"^s2^"<B>)</B>") ;
       []
 | Spat_list (patl,s1,s2) ->
       emit_string (s1^"<B>[</B>") ;
       let newdefs = emit_list (produce_pattern env) patl in
         emit_string (s2^"<B>]</B>") ;
         newdefs
 | Spat_withtype (pat,typ,s1,s2,s3) ->
       emit_string (s1^"<B>(</B>") ;
       let newdefs = produce_pattern env pat in
        emit_string (s2^"<B>:</B>") ;
        produce_type env typ ;
        emit_string (s3^"<B>)</B>") ;
        newdefs
 | Spat_record (patlabl,s1,s2) ->
       emit_string (s1^"<B>{</B>") ;
       let newdefs = emit_list (produce_patlabel env) patlabl in
        emit_string (s2^"<B>}</B>") ;
        newdefs
 | Spat_paren (pat,s1,s2) ->
       emit_string (s1^"<B>(</B>") ;
       let newdefs = produce_pattern env pat in
        emit_string (s2^"<B>)</B>") ;
        newdefs
 | Spat_charinterv (ch1,ch2,s1,s2,s3) ->
       emit_string s1 ;
       emit_string "`" ;
       emit_string ch1 ;
       emit_string "`" ;
       emit_string s2 ;
       emit_string "<B>..</B>" ;
       emit_string s3 ;
       emit_string "`" ;
       emit_string ch2 ;
       emit_string "`";
       []

(* FOR NON RECURSIVE BINDINGS *)
and produce_nonrec_binding env = function
   Bind_pattern (pat,e,s) ->
       (* USED FOR let f = ... *)
       let newdefs = produce_pattern env pat in
        emit_string (s^"<B>=</B>") ;
        produce_expr env e ;
        newdefs
 | Bind_ident (id,spatl,e,s) ->
       (* USED FOR let f x y z = ... *)
       let newdefs1 = produce_and_store_ident_value env id in
        let newdefs2 = emit_simplelist (produce_spattern env) spatl in
         emit_string (s^"<B>=</B>") ;
         produce_expr (newdefs2@env) e ;
         newdefs1

(* FOR RECURSIVE BINDINGS *)
and produce_rec_binding env = function
   Bind_pattern (pat,e,s) ->
       (* USED FOR let f = ... *)
       let newdefs = produce_pattern env pat in
        emit_string (s^"<B>=</B>") ;
        produce_expr (newdefs@env) e ;
        newdefs
 | Bind_ident (id,spatl,e,s) ->
       (* USED FOR let f x y z = ... *)
       let newdefs1 = produce_and_store_ident_value env id in
        let newdefs2 = emit_simplelist (produce_spattern env) spatl in
         emit_string (s^"<B>=</B>") ;
         produce_expr (newdefs2@newdefs1@env) e ;
         newdefs1

and produce_recordelem env (eid,exp,str) =
 produce_and_mark_extid_value env eid ;
 emit_string (str^"<B>=</B>") ;
 produce_expr env exp

and produce_streamexprcomp env = function
   Sec_expr e ->
      produce_expr env e
 | Sec_quote (e,s) ->
      emit_string (s^"<B>'</B>") ;
      produce_expr env e

and produce_sexpr env = function
   Sexp_structcst atom ->
       produce_atomcst env atom
 | Sexp_extident eid ->
       produce_and_mark_extid_value env eid
 | Sexp_list (expl,s1,s2) ->
       emit_string (s1^"<B>[</B>") ;
       do_seplist (produce_expr env) expl ;
       emit_string (s2^"<B>]</B>")
 | Sexp_vector (expl,s1,s2) ->
       emit_string (s1^"<B>[|</B>") ;
       do_seplist (produce_expr env) expl ;
       emit_string (s2^"<B>|]</B>")
 | Sexp_stream (strexpcompl,s1,s2) ->
       emit_string (s1^"<B>[<</B>") ;
       do_seplist (produce_streamexprcomp env) strexpcompl ;
       emit_string (s2^"<B>>]</B>")
 | Sexp_withtype (exp,typ,s1,s2,s3) ->
       emit_string (s1^"<B>(</B>") ;
       produce_expr env exp ;
       emit_string (s2^"<B>:</B>") ;
       produce_type env typ ;
       emit_string (s3^"<B>)</B>")
 | Sexp_paren (expopt,s1,s2) ->
       emit_string (s1^"<B>(</B>") ;
       (match expopt with None -> () | Some e -> produce_expr env e) ;
       emit_string (s2^"<B>)</B>")
 | Sexp_begin(expopt,s1,s2) ->
       emit_string (s1^"<B>begin</B>") ;
       (match expopt with None -> () | Some e -> produce_expr env e) ;
       emit_string (s2^"<B>end</B>")
 | Sexp_record (recelems,s1,s2) ->
       emit_string (s1^"<B>{</B>") ;
       do_seplist (produce_recordelem env) recelems ;
       emit_string (s2^"<B>}</B>")
 | Sexp_prefix (s1,sexp,s2) ->
       emit_string (s2^s1) ;
       produce_sexpr env sexp
 | Sexp_dot (sexp,eid,s) ->
       (* WE DON'T ACTUALLY HANDLE RECORD FIELDS *)
       produce_sexpr env sexp ;
       emit_string (s^".") ;
       produce_extid env eid
 | Sexp_dotlparen (sexp,exp,s1,s2) ->
       produce_sexpr env sexp ;
       emit_string (s1^"<B>.(</B>") ;
       produce_expr env exp ;
       emit_string (s2^"<B>)</B>")
 | Sexp_dotlbracket (sexp,exp,s1,s2) ->
       produce_sexpr env sexp ;
       emit_string (s1^"<B>.[</B>") ;
       produce_expr env exp ;
       emit_string (s2^"<B>]</B>")

and produce_patact env (p,a) =
 let newdefs = produce_pattern env p in
  produce_action (newdefs@env) a ;
  ()

and produce_funmatch env (spatl,act) =
 emit_simplelist (produce_spattern env) spatl ;
 produce_action env act

and produce_action env = function
   Act_direct (e,s) ->
      emit_string (s^"<B>-&gt;</B>") ;
      produce_expr env e
 | Act_when (e1,e2,s1,s2) ->
      emit_string (s1^"<B>when</B>") ;
      produce_expr env e1 ;
      emit_string (s2^"<B>-&gt;</B>") ;
      produce_expr env e2

and produce_funstreamelems env (streampat,e,s) =
 let newdefs = produce_streampat env streampat in
  emit_string (s^"<B>-&gt;</B>") ;
  (* DEFS DON'T GO OUT *)
  produce_expr (newdefs@env) e

and produce_streampatcomp env = function
   Spc_sexprspat (sexp,spat) ->
      produce_sexpr env sexp ;
      produce_spattern env spat
 | Spc_quote (pat,s) ->
      emit_string (s^"<B>'</B>") ;
      produce_pattern env pat
 | Spc_ident (s1,s2) ->
      let mark = newmarker () in
       emit_string s2 ;
       emit_aname (s1^(string_of_int mark)) s1 ;
       [(s1,Obj_value,mark)]
 | Spc_semi (spc,s) ->
      emit_string (s^"<B>;</B>") ;
      produce_streampatcomp env spc

and produce_streampat env = function
   Strp_full (spcl,s1,s2) ->
       emit_string (s1^"<B>[&lt;</B>") ;
       let newdefs = emit_list (produce_streampatcomp env) spcl in
        emit_string (s2^"<B>&gt;]</B>") ;
        newdefs
 | Strp_empty (s1,s2) ->
       emit_string (s1^"<B>[&lt;</B>"^s2^"<B>&gt;]</B>") ;
       []

and produce_expr env = function
   Expr_sexpr sexp ->
       produce_sexpr env sexp
 | Expr_sexprlst sexpl ->
       do_list (produce_sexpr env) sexpl
 | Expr_exprcomma expl ->
       do_seplist (produce_expr env) expl
 | Expr_substr1 (s1,exp,s2) ->
       emit_string (s2^s1) ;
       produce_expr env exp
 | Expr_not (exp,s) ->
       emit_string (s^"<B>not</B>") ;
       produce_expr env exp
 | Expr_lessminus (id,exp,s) ->
       produce_and_mark_ident_value env id ;
       emit_string (s^"-") ;
       produce_expr env exp
 | Expr_infix (s1,e1,e2,s2) ->
       produce_expr env e1 ;
       emit_string s2 ;
       (* MADE BY HAND BECAUSE s1 IS A STRING, NOT AN IDENTIFIER *)
       (match is_defined s1 Obj_value env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1) ;
       produce_expr env e2
 | Expr_substr2 (s1,e1,e2,s2) ->
       produce_expr env e1 ;
       emit_string s2 ;
       (match is_defined s1 Obj_value env with
          Some (mark) -> emit_href_local (s1^(string_of_int mark)) s1
        | None -> emit_string s1) ;
       produce_expr env e2
 | Expr_star (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>*</B>") ;
       produce_expr env e2
 | Expr_coloncolon (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>::</B>") ;
       produce_expr env e2
 | Expr_equal (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>=</B>") ;
       produce_expr env e2
 | Expr_equalequal (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>==</B>") ;
       produce_expr env e2
 | Expr_ampersand (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>&amp;</B>") ;
       produce_expr env e2
 | Expr_amperamper (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>&amp;&amp;</B>") ;
       produce_expr env e2
 | Expr_or (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>or</B>") ;
       produce_expr env e2
 | Expr_barbar (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>||</B>") ;
       produce_expr env e2
 | Expr_dotlessminus (sexp,eid,exp,s1,s2) ->
       produce_sexpr env sexp ;
       emit_string (s1^"<B>.</B>") ;
       produce_extid env eid ;  (* WE DON'T ACTUALLY HANDLE RECORD FIELDS *)
       emit_string (s2^"<B>&lt;-</B>") ;
       produce_expr env exp
 | Expr_dotlparen (sexp,e1,e2,s1,s2,s3) ->
       produce_sexpr env sexp ;
       emit_string (s1^"<B>.(</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>)</B>") ;
       emit_string (s3^"<B>&lt;-</B>") ;
       produce_expr env e2
 | Expr_dotlbracket(sexp,e1,e2,s1,s2,s3) ->
       produce_sexpr env sexp ;
       emit_string (s1^"<B>.[</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>]</B>") ;
       emit_string (s3^"<B>&lt;-</B>") ;
       produce_expr env e2
 | Expr_colonequal (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>:=</B>") ;
       produce_expr env e2
 | Expr_ifthenelse (e1,e2,e3,s1,s2,s3) ->
       emit_string (s1^"<B>if</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>then</B>") ;
       produce_expr env e2 ;
       emit_string (s2^"<B>else</B>") ;
       produce_expr env e3
 | Expr_ifthen (e1,e2,s1,s2) ->
       emit_string (s1^"<B>if</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>then</B>") ;
       produce_expr env e2
 | Expr_while (e1,expopt,s1,s2,s3) ->
       emit_string (s1^"<B>while</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>do</B>") ;
       (match expopt with None -> () | Some e -> produce_expr env e) ;
       emit_string (s3^"<B>done</B>")
 | Expr_forup (id,e1,e2,expopt,s1,s2,s3,s4,s5) ->
       emit_string (s1^"<B>for</B>") ;
       let newdefs = produce_and_store_ident_value env id in
        emit_string (s2^"<B>=</B>") ;
        produce_expr (newdefs@env) e1 ;
        emit_string (s3^"<B>to</B>") ;
        produce_expr (newdefs@env) e2 ;
        emit_string (s4^"<B>do</B>") ;
        (match expopt with None -> ()
                         | Some e -> produce_expr (newdefs@env) e) ;
        emit_string (s5^"<B>done</B>")
 | Expr_fordown (id,e1,e2,expopt,s1,s2,s3,s4,s5) ->
       emit_string (s1^"<B>for</B>") ;
       let newdefs = produce_and_store_ident_value env id in
        emit_string (s2^"<B>=</B>") ;
        produce_expr (newdefs@env) e1 ;
        emit_string (s3^"<B>downto</B>") ;
        produce_expr (newdefs@env) e2 ;
        emit_string (s4^"<B>do</B>") ;
        (match expopt with None -> ()
                         | Some e -> produce_expr (newdefs@env) e) ;
        emit_string (s5^"<B>done</B>")
 | Expr_semisemi (e1,e2,s) ->
       produce_expr env e1 ;
       emit_string (s^"<B>;</B>") ;
       produce_expr env e2
 | Expr_semi (exp,s) ->
       produce_expr env exp ;
       emit_string (s^"<B>;</B>")
 | Expr_match (e1,patactl,s1,s2,s3) ->
       emit_string (s1^"<B>match</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>with</B>") ;
       emit_string s3 ;
       do_seplist (produce_patact env) patactl
 | Expr_matchstream (e1, patesl, s1,s2,s3) ->
       emit_string (s1^"<B>match</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>with</B>") ;
       emit_string s3 ;
       do_seplist (produce_funstreamelems env) patesl
 | Expr_let (bndsl,e,s1,s2) ->
       emit_string (s1^"<B>let</B>") ;
       let newdefs = emit_list (produce_nonrec_binding env) bndsl in
        emit_string (s2^"<B>in</B>") ;
        produce_expr (newdefs@env) e
 | Expr_letrec (bndsl,e,s1,s2,s3) ->
       emit_string (s1^"<B>let</B>"^s2^"<B>rec</B>") ;
       let newdefs = emit_list (produce_rec_binding env) bndsl in
        emit_string (s3^"<B>in</B>") ;
        produce_expr (newdefs@env) e
 | Expr_fun (funmatch,s1,s2) ->
       emit_string (s1^"<B>fun</B>") ;
       emit_string s2 ;
       do_seplist (produce_funmatch env) funmatch
 | Expr_function (patactl,s1,s2) ->
       emit_string (s1^"<B>function</B>") ;
       emit_string s2 ;
       do_seplist (produce_patact env) patactl
 | Expr_functionstream (stpatesl,s1,s2) ->
       emit_string (s1^"<B>function</B>") ;
       emit_string s2 ;
       do_seplist (produce_funstreamelems env) stpatesl
 | Expr_try (e1,patactl,s1,s2,s3) ->
       emit_string (s1^"<B>try</B>") ;
       produce_expr env e1 ;
       emit_string (s2^"<B>with</B>") ;
       emit_string s3 ;
       do_seplist (produce_patact env) patactl
 | Expr_where (e,bdnsl,s) ->
       (* NOT REALLY CORRECT BECAUSE ANCHORAGE ARE NOT HANDLED *)
       produce_expr env e ;
       emit_string (s^"<B>where</B>") ;
       emit_list (produce_nonrec_binding env) bdnsl ;
       (* WE THROW BINDINGS FOR THE MOMENT. TO SEE I LATER VERSION *)
       ()
 | Expr_whererec (e,bdnsl,s1,s2) ->
       (* NOT RALLY CORRECT BECAUSE ANCHORAGE ARE NOT HANDLED *)
       produce_expr env e ;
       emit_string (s1^"<B>where</B>") ;
       emit_string (s2^"<B>rec</B>") ;
       emit_list (produce_rec_binding env) bdnsl ;
       (* WE THROW BINDINGS FOR THE MOMENT. TO SEE I LATER VERSION *)
       ()

and produce_mutable env = function
   Mut_yes s ->
      emit_string (s^"<B>mutable</B>")
 | Mut_no ->
      ()

and produce_constr env = function
   Tc_arityn (s1,mut,typ,s2,s3) ->
     emit_string (s2^s1) ;
     emit_string (s3^"<B>of</B>") ;
     produce_mutable env mut ;
     produce_type env typ
 | Tc_arity0 (s1,s2) ->
    (* WE DON'T SEARCH FOR ANCHORAGE *)
     emit_string (s2^s1)

and produce_excconstr env = function
   Tc_arityn (s1,mut,typ,s2,s3) ->
     let mark = newmarker () in
      emit_string s2 ;
      emit_aname (s1^(string_of_int mark)) s1 ;
      emit_string (s3^"<B>of</B>") ;
      produce_mutable env mut ;
      produce_type env typ ;
      [(s1,Obj_value,mark)]
 | Tc_arity0 (s1,s2) ->
     let mark = newmarker () in
      emit_string s2 ;
      emit_aname (s1^(string_of_int mark)) s1 ;
      [(s1,Obj_value,mark)]
     

and produce_tparam env = function
   Tpar_paren (lst,s1,s2) ->
       emit_string (s1^"(") ;
       emit_list (fun (s1,s2,s3)->emit_string (s2^"<B>'</B>"^s3^s1);[]) lst ;
       emit_string (s2^")") ;
       []
 | Tpar_var (s1,s2,s3) ->
       emit_string (s2^"<B>'</B>"^s3^s1) ;
       []
 | Tpar_none ->
       []

and produce_labeldecl env (mut,s1,typ,s2,s3) =
 produce_mutable env mut ;
 emit_string (s2^s1) ;
 emit_string (s3^"<B>:</B>") ;
 produce_type env typ ;

and produce_typedef env = function
   Tdef_abstract -> ()
 | Tdef_sum (constrl,s1,s2) ->
       emit_string (s1^"<B>=</B>") ;
       emit_string s2 ;
       do_seplist (produce_constr env) constrl
 | Tdef_record (recelms,s1,s2,s3) ->
       emit_string (s1^"<B>=</B>") ;
       emit_string (s2^"<B>{</B>") ;
       do_seplist (produce_labeldecl env) recelms ;
       emit_string (s3^"<B>}</B>")
 | Tdef_abrev (typ,s) ->
       emit_string (s^"<B>==</B>") ;
       produce_type env typ

and produce_directive env ((s1,s2),(s3,s4)) =
  emit_string ("<FONT COLOR=\"#"^(!directivecolor)^"\">") ;
  emit_string (s2^s1^s4) ;
  emit_string "</FONT>" ;
  emit_string ("<FONT COLOR=\"#"^(!directiveargcolor)^"\">") ;
  emit_string ("\""^s3^"\"") ;
  emit_string "</FONT>"

and produce_typdeclcomp env (typepar,s1,typdef,s2) =
 produce_tparam env typepar ;
 let mark = newmarker () in
  emit_string s2 ;
  emit_aname (s1^(string_of_int mark)) s1 ;
  let newdefs = [(s1,Obj_type,mark)] in
   produce_typedef (newdefs@env) typdef ;
   newdefs

and produce_implementation env = function
    Imp_expr (e,s) ->
        produce_expr !env e ;
        emit_string (s^"<B>;;</B>") ;
        emit_flush ()
  | Imp_let (bndsl,s1,s2) ->
        emit_string (s1^"<B>let</B>") ;
        let newdefs = emit_list (produce_nonrec_binding !env) bndsl in
         emit_string (s2^"<B>;;</B>") ;
         env := newdefs@(!env) ;
         emit_flush ()
  | Imp_letrec (bndsl,s1,s2,s3) ->
        emit_string (s1^"<B>let</B>"^s2^"<B>rec</B>") ;
        let newdefs = emit_list (produce_rec_binding !env) bndsl in
        emit_string (s3^"<B>;;</B>") ;
        env := newdefs@(!env) ;
        emit_flush ()
  | Imp_type (l,s1,s2) ->
        emit_string (s1^"<B>type</B>") ;
        let newdefs = emit_list (produce_typdeclcomp !env) l in
         emit_string (s2^"<B>;;</B>") ;
         env := newdefs@(!env) ;
         emit_flush ()
  | Imp_exception (constrl,s1,s2) ->
        emit_string (s1^"<B>exception</B>") ;
        let newdefs = emit_list (produce_excconstr !env) constrl in
         emit_string (s2^"<B>;;</B>") ;
         env := newdefs@(!env) ;
         emit_flush ()
  | Imp_directive (dircomp,s5,s6) ->
       emit_string (s5^"<B>#</B>") ;
       produce_directive !env dircomp ;
       emit_string (s6^"<B>;;</B>") ;
       emit_flush ()
;;

