{
(* REMARK : HTML TRANSLATION ID HANDLED BY A PREPROC FOR COMMENTS *)
(* AND DIRECTLY IN LEX PRODUCTION FOR STRINGS. MAY BE IT COULD BE *)
(* BETTER TO HANDLE ALL IN THE SAME WAY. SEE...                   *)
#open "parser";;
#open "misc" ;;


type lexical_error =
    Illegal_character
  | Unterminated_comment
  | Bad_char_constant
  | Unterminated_string;;
exception Lexical_error of lexical_error * int * int;;

(* Pre-values for tokens (need to be added inter_token symbols *)
type Token_firstvalue = T_and|T_as|T_begin
 |T_do|T_done|T_downto|T_else
 |T_end|T_exception|T_for|T_fun
 |T_function|T_if|T_in|T_let
 |T_match|T_mutable|T_not|T_of
 |T_or|T_pref|T_rec|T_then
 |T_to|T_try|T_type|T_value
 |T_when|T_where|T_while|T_with
 |T_inf3 of string|T_inf4 of string
;;

(* Text between lexical symbols *)
let inter_token_text = ref (create_string 256) ;;
inter_token_text := "" ;;

(* For nested comments *)
let comment_depth = ref 0;;

(* The table of keywords *)
let keyword_table = (hashtbl__new 149 : (string, Token_firstvalue) hashtbl__t)
;;

do_list (fun (str,tok) -> hashtbl__add keyword_table str tok) [
  "and", T_and;
  "as", T_as;
  "begin", T_begin;
  "do", T_do;
  "done", T_done;
  "downto", T_downto;
  "else", T_else;
  "end", T_end;
  "exception", T_exception;
  "for", T_for;
  "fun", T_fun;
  "function", T_function;
  "if", T_if;
  "in", T_in;
  "let", T_let;
  "match", T_match;
  "mutable", T_mutable;
  "not", T_not;
  "of", T_of;
  "or", T_or;
  "prefix", T_pref;
  "rec", T_rec;
  "then", T_then;
  "to", T_to;
  "try", T_try;
  "type", T_type;
  "value", T_value;
  "when", T_when;
  "where", T_where;
  "while", T_while;
  "with", T_with;

  "quo", T_inf3("quo");
  "mod", T_inf3("mod");
  "land", T_inf3("land");
  "lor", T_inf3("lor");
  "lxor", T_inf3("lxor");
  "lsl", T_inf4("lsl");
  "lsr", T_inf4("lsr");
  "asr", T_inf4("asr")
];;



(* To buffer string literals *)
let initial_string_buffer = create_string 256;;
let string_buff = ref initial_string_buffer;;
let string_index = ref 0;;

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0;
  ()
;;

let store_string_char c =
  if !string_index >= string_length (!string_buff) then begin
    let new_buff = create_string (string_length (!string_buff) * 2) in
      blit_string (!string_buff) 0 new_buff 0 (string_length (!string_buff));
      string_buff := new_buff
  end;
  set_nth_char (!string_buff) (!string_index) c;
  incr string_index
;;

let get_stored_string () =
  let s = sub_string (!string_buff) 0 (!string_index) in
    string_buff := initial_string_buffer;
    s
;;

(* To translate COMMENTS with &, <, > to HTML tags strings comments *)
(* ALSO USED  FOR infix AND substr *)
let comment_for_html s =
 let max_i = string_length s in
  let rec parsestr i =
   if i < max_i
    then
     begin
      let ns =
       match sub_string s i 1 with
          "&" -> "&amp;"
        | "<" -> "&lt;"
        | ">" -> "&gt;"
        | others -> others
      in
       ns^(parsestr (i+1))
     end
  else ""
 in parsestr 0
;;


let token_for_final_value tk =
 let tmp = !inter_token_text in
  inter_token_text := "" ;
  match tk with
    T_and->AND(tmp)|T_as->AS(tmp)|T_begin->BEGIN(tmp)|T_do->DO(tmp)
   |T_done->DONE(tmp)|T_downto->DOWNTO(tmp)|T_else->ELSE(tmp)
   |T_end->END(tmp)|T_exception->EXCEPTION(tmp)|T_for->FOR(tmp)
   |T_fun->FUN(tmp)|T_function->FUNCTION(tmp)|T_if->IF(tmp)
   |T_in->IN(tmp)|T_let->LET(tmp)|T_match->MATCH(tmp)
   |T_mutable->MUTABLE(tmp)|T_not->NOT(tmp)|T_of->OF(tmp)
   |T_or->OR(tmp)|T_pref->PREF(tmp)|T_rec->REC(tmp)|T_then->THEN(tmp)
   |T_to->TO(tmp)|T_try->TRY(tmp)|T_type->TYPE(tmp)|T_value->VALUE(tmp)
   |T_when->WHEN(tmp)|T_where->WHERE(tmp)|T_while->WHILE(tmp)
   |T_with->WITH(tmp)|T_inf3(s)->INFIX3(s,tmp)|T_inf4(s)->INFIX4(s,tmp)
;;


}


rule main = parse
    [` ` `\010` `\013` `\009` `\012`] +
      { inter_token_text := !inter_token_text ^ (get_lexeme lexbuf) ;
        main lexbuf }
  | [`A`-`Z` `a`-`z` `\192`-`\214` `\216`-`\246` `\248`-`\255`  ]
    ( `_` ? [`A`-`Z` `a`-`z` `\192`-`\214` `\216`-`\246` `\248`-`\255`  `'` (*'*) `0`-`9` ] ) *
      { let s = get_lexeme lexbuf in
          try
            token_for_final_value (hashtbl__find keyword_table s)
          with Not_found ->
            let tmp = !inter_token_text in
              inter_token_text := "" ;
              IDENT (s,tmp) }
    | [`0`-`9`]+
    | `0` [`x` `X`] [`0`-`9` `A`-`F` `a`-`f`]+
    | `0` [`o` `O`] [`0`-`7`]+
    | `0` [`b` `B`] [`0`-`1`]+
      { let tmp = !inter_token_text in
        inter_token_text := "" ;
        INT (int_of_string(get_lexeme lexbuf),tmp) }
  | [`0`-`9`]+ (`.` [`0`-`9`]*)? ([`e` `E`] [`+` `-`]? [`0`-`9`]+)?
      { let tmp = !inter_token_text in
        inter_token_text := "" ;
        FLOAT (float_of_string(get_lexeme lexbuf),tmp) }
  | "\""
      { reset_string_buffer();
        let string_start = lexbuf.lex_start_pos + lexbuf.lex_abs_pos in
        begin try
          string lexbuf
        with Lexical_error(Unterminated_string, _, string_end) ->
          raise(Lexical_error(Unterminated_string, string_start, string_end))
        end;
        lexbuf.lex_start_pos <- string_start - lexbuf.lex_abs_pos;
        let tmp = !inter_token_text in
          inter_token_text := "" ;
          STRING (get_stored_string(),tmp) }
  | "`"
      { let char_start = lexbuf.lex_start_pos + lexbuf.lex_abs_pos in
        let c = char lexbuf in
        lexbuf.lex_start_pos <- char_start - lexbuf.lex_abs_pos;
        let tmp = !inter_token_text in
          inter_token_text := "" ;
          CHAR (c,tmp) }
  | "(*"
      { (* WE MEMORIZE COMMENTS WITH HTLM EMPHASIS AT BEGINNING *)
        inter_token_text:=!inter_token_text^"<EM><FONT COLOR=\"#"^ !commentcolor ^"\">(*" ;
        let comment_start = lexbuf.lex_start_pos + lexbuf.lex_abs_pos in
        comment_depth := 1;
        begin try
          comment lexbuf
        with Lexical_error(Unterminated_comment, _, comment_end) ->
          raise(Lexical_error(Unterminated_comment,
                              comment_start, comment_end))
        end;
        main lexbuf }
  | "#" { let tmp = !inter_token_text in inter_token_text := "";
          SHARP (tmp) }
  | "&" { let tmp = !inter_token_text in inter_token_text := "";
          AMPERSAND (tmp) }
  | "'" { let tmp = !inter_token_text in inter_token_text := "";
          QUOTE(tmp) }
  | "(" { let tmp = !inter_token_text in inter_token_text := "";
          LPAREN(tmp) }
  | ")" { let tmp = !inter_token_text in inter_token_text := "";
          RPAREN(tmp) }
  | "*" { let tmp = !inter_token_text in inter_token_text := "";
          STAR(tmp) }
  | "," { let tmp = !inter_token_text in inter_token_text := "";
          COMMA(tmp) }
  | "->" { let tmp = !inter_token_text in inter_token_text := "";
           MINUSGREATER(tmp) }
  | "." { let tmp = !inter_token_text in inter_token_text := "";
          DOT(tmp) }
  | ".." { let tmp = !inter_token_text in inter_token_text := "";
           DOTDOT(tmp) }
  | ".(" { let tmp = !inter_token_text in inter_token_text := "";
           DOTLPAREN(tmp) }
  | ".[" { let tmp = !inter_token_text in inter_token_text := "";
           DOTLBRACKET(tmp) }
  | ":" { let tmp = !inter_token_text in inter_token_text := "";
          COLON(tmp) }
  | "::" { let tmp = !inter_token_text in inter_token_text := "";
           COLONCOLON(tmp) }
  | ":=" { let tmp = !inter_token_text in inter_token_text := "";
           COLONEQUAL(tmp) }
  | ";" { let tmp = !inter_token_text in inter_token_text := "";
          SEMI(tmp) }
  | ";;" { let tmp = !inter_token_text in inter_token_text := "";
           SEMISEMI(tmp) }
  | "<-" { let tmp = !inter_token_text in inter_token_text := "";
           LESSMINUS(tmp) }
  | "=" { let tmp = !inter_token_text in inter_token_text := "";
          EQUAL(tmp) }
  | "==" { let tmp = !inter_token_text in inter_token_text := "";
           EQUALEQUAL(tmp) }
  | ">]" { let tmp = !inter_token_text in inter_token_text := "";
           GREATERRBRACKET(tmp) }
  | "[" { let tmp = !inter_token_text in inter_token_text := "";
          LBRACKET(tmp) }
  | "[|" {let tmp = !inter_token_text in inter_token_text := "";
          LBRACKETBAR(tmp) }
  | "[<" { let tmp = !inter_token_text in inter_token_text := "";
           LBRACKETLESS(tmp) }
  | "]" { let tmp = !inter_token_text in inter_token_text := "";
          RBRACKET(tmp) }
  | "_" { let tmp = !inter_token_text in inter_token_text := "";
          UNDERSCORE(tmp) }
  | "__" { let tmp = !inter_token_text in inter_token_text := "";
           UNDERUNDER(tmp) }
  | "{" { let tmp = !inter_token_text in inter_token_text := "";
          LBRACE(tmp) }
  | "|" { let tmp = !inter_token_text in inter_token_text := "";
          BAR(tmp) }
  | "|]" { let tmp = !inter_token_text in inter_token_text := "";
           BARRBRACKET(tmp) }
  | "}" { let tmp = !inter_token_text in inter_token_text := "";
          RBRACE(tmp) }
  | "&&" { let tmp = !inter_token_text in inter_token_text := "";
           AMPERAMPER(tmp) }
  | "||" { let tmp = !inter_token_text in inter_token_text := "";
           BARBAR(tmp) }
  | "!="    { let tmp = !inter_token_text in inter_token_text := "";
              INFIX0 ("!=",tmp) }
  | "-"     { let tmp = !inter_token_text in inter_token_text := "";
              SUBTRACTIVE ("-",tmp) }
  | "-."    { let tmp = !inter_token_text in inter_token_text := "";
              SUBTRACTIVE ("-.",tmp) }
  | [ `!` `?` ] [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              PREFIX(comment_for_html(get_lexeme lexbuf),tmp) }
  | [ `=` `<` `>` `|` `&` `~` `$` ] [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              INFIX0(comment_for_html(get_lexeme lexbuf),tmp) }
  | [ `@` `^` ] [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              INFIX1(comment_for_html(get_lexeme lexbuf),tmp) }
  | [ `+` `-` ] [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              INFIX2(comment_for_html(get_lexeme lexbuf),tmp) }
  | "**" [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              INFIX4(comment_for_html(get_lexeme lexbuf),tmp) }
  | [ `*` `/` `%` ] [ `!` `$` `%` `&` `*` `+` `-` `.` `/` `:` `<` `=` `>` `?` `@` `^` `|` `~`  ] *
            { let tmp = !inter_token_text in inter_token_text := "";
              INFIX3(comment_for_html(get_lexeme lexbuf),tmp) }
  | eof  { let tmp = !inter_token_text in inter_token_text := "";
           EOF(tmp) }
  | _
      { raise (Lexical_error(Illegal_character,
                            get_lexeme_start lexbuf, get_lexeme_end lexbuf)) }

and comment = parse
    "(*"
      { (* WE KEEP ON MEMORIZING COMMENTS *)
        inter_token_text:=!inter_token_text^"(*" ;
        comment_depth := succ !comment_depth; comment lexbuf }
  | "*)"
      { (* WE MUST KEEP THE "*)" DELIMITER EVEN IF IT IS A NESTED COMMENT *)
        inter_token_text:=!inter_token_text^"*)" ;
        comment_depth := pred !comment_depth;
        if !comment_depth > 0
         then comment lexbuf
         else (* IF IT'S THE END OF COMMENT WE END HTML EMPHASIS *)
          inter_token_text:=!inter_token_text^"</FONT></EM>" }
  | "\""
      { (* WE KEEP THER SAME BEHAVIOUR THAN CAML WHEN WE FIND A *)
        (* DOUBLE QUOTE IN COMMENT... WAIT FOR CLOSING DOUBLE QUOTE *)
        (* OTHERWISE RAISING Unterminated_string *)
        reset_string_buffer();
        let string_start = lexbuf.lex_start_pos + lexbuf.lex_abs_pos in
        begin try
          string lexbuf
        with Lexical_error(Unterminated_string, _, string_end) ->
          raise(Lexical_error(Unterminated_string, string_start, string_end))
        end;
        comment lexbuf }
  | "``"
      { inter_token_text:=!inter_token_text^"``" ; comment lexbuf }
  | "`" [^ `\\` `\``] "`"
      { (* QUOTED CHAR EXPRESSION IN COMMENT *)
        inter_token_text:=!inter_token_text^(comment_for_html (get_lexeme lexbuf)) ;
        comment lexbuf }
  | "`" `\\` [`\\` `\`` `n` `t` `b` `r`] "`"
      { (* QUOTED DELIMITER EXPRESSION IN COMMENT *)
        inter_token_text:=!inter_token_text^(get_lexeme lexbuf) ;
        comment lexbuf }
  | "`" `\\` [`0`-`9`] [`0`-`9`] [`0`-`9`] "`"
      { (* QUOTED ASCII CHAR EXPRESSION IN COMMENT *)
        inter_token_text:=!inter_token_text^(get_lexeme lexbuf) ;
        comment lexbuf }
  | eof 
      { raise(Lexical_error
                (Unterminated_comment, 0, get_lexeme_start lexbuf)) }
  | _
      { inter_token_text:=!inter_token_text^(comment_for_html (get_lexeme lexbuf)) ;
        comment lexbuf }


and char = parse
    [^ `\``] "`"
      { (* WE CAN USE char_for_read FOR TRANSLATING CHAR TO STRING *)
        (* HERE BECAUSE WE'RE SUR THERE IS NO SPECIAL CHARACTERS *)
        (* AT THIS POINT. *)
        (* WE THEREFORE MUST TRANSLATE HTML TAGS *)
        match (get_lexeme_char lexbuf 0) with
           `&` -> "&amp;"
         | `<` -> "&lt;"
         | `>` -> "&gt;"
         | other -> char_for_read (get_lexeme_char lexbuf 0) }
  | `\\` [`n` `t` `b` `r`] "`"
      { (* WE MUST BE CAREFUL BECAUSE slash AND backquote WILL BE *)
        (* INTERPRETED BY *)
        (* char_for_read SO WE MUST KEEP IT BEFORE *)
        ("\\"^(char_for_read (get_lexeme_char lexbuf 1))) }
  | `\\` [`\\`] "`"
      { "\\\\" }
  | `\\` [`\``] "`"
      { "\\`" }
  | `\\` [`0`-`9`] [`0`-`9`] [`0`-`9`] "`"
      { (* IDEM *)
        ("\\"^
         (char_for_read (get_lexeme_char lexbuf 1))^
         (char_for_read (get_lexeme_char lexbuf 2))^
         (char_for_read (get_lexeme_char lexbuf 3)))}
  | [^ `\``] * ("`" | eof )
      { raise (Lexical_error(Bad_char_constant,
                            get_lexeme_start lexbuf - 1,
                            get_lexeme_end lexbuf)) }

and string = parse
    `"` (*"'"'*)
      { () }
  | `\\` ("\010" | "\013" | "\013\010")          (* [` ` `\009`] * *)
      { (* WE CATCH LINE CONTINUATIONS *)
        store_string_char `\\` ;
        store_string_char(get_lexeme_char lexbuf 1);
        string lexbuf }
  | `\\` [`\\` `"` (*"'"'*) `n` `t` `b` `r`]
      { (* WE DON'T INTERPRET SEQUENCE *)
        store_string_char `\\` ;
        store_string_char (get_lexeme_char lexbuf 1) ;
        string lexbuf }
  | `\\` [`0`-`9`] [`0`-`9`] [`0`-`9`]
      { (* WE PREVENT FROM INTERPRETING ASCII CODE SEQUENCE *)
        (* BY BLASTING IT DIRECTLY IN THE LEXING BUFFER *)
        store_string_char `\\` ;
        store_string_char (get_lexeme_char lexbuf 1) ;
        store_string_char (get_lexeme_char lexbuf 2) ;
        store_string_char (get_lexeme_char lexbuf 3) ;
        string lexbuf }
  | eof 
      { raise (Lexical_error
                (Unterminated_string, 0, get_lexeme_start lexbuf)) }
  | _
      { (* WR CONVERT <, > AND & TO CORRECT HTML TAGS *)
        begin
        match get_lexeme_char lexbuf 0 with
           `&` -> store_string_char `&` ;
                  store_string_char `a` ;
                  store_string_char `m` ;
                  store_string_char `p` ;
                  store_string_char `;`
         | `<` ->store_string_char `&` ;
                  store_string_char `l` ;
                  store_string_char `t` ;
                  store_string_char `;`
         | `>` ->store_string_char `&` ;
                  store_string_char `g` ;
                  store_string_char `t` ;
                  store_string_char `;`
         | other -> store_string_char other
        end ;
        string lexbuf }
;;
