#open "lexer";;
#open "parser";;
#open "analyser";;
#open "misc" ;;


(* PARSE AN GENERATE 1 FILE CONVERSION ON STDOUT *)
let convert_file filename =
 try
  let in_channel = open_in filename in
   misc__current_out_channel := if !misc__stdoutflag then stdout
                          else open_out (filename^".html") ;
   begin
   try
    let lexbuf = lexing__create_lexer_channel in_channel in
     (* IF RESULTS ARE IN SEPARATE FILE, WE CAN RESET COUNTER *)
     (* BECAUSE THERE WONT BE ANY CLASHING ANCHORS *)
     if !misc__stdoutflag then resetmarker () ;
     (if !misc__headerflag then emit_string "<HTML>\n<BODY>\n<PRE>\n") ;
     while true do
      let tree = parser__Implementation lexer__main lexbuf in
       produce_implementation envdef tree
     done
   with
      treetype__End_of_file ->
        if !misc__headerflag then
         begin
         emit_string "\n</PRE>\n</BODY>\n</HTML>\n" ;
         emit_flush ()
         end
    | parsing__Parse_error ->
        (* let pos1 = lexing__get_lexeme_start lexbuf in
           let pos2 = lexing__get_lexeme_end lexbuf in *)
        prerr_string "Syntax error.\n";
        emit_flush () ;
        exit (-1)
    | lexer__Lexical_error (errcode, pos1, pos2) ->
        (match errcode with
           lexer__Illegal_character ->
            prerr_string "Illegal character.\n"
         | lexer__Unterminated_comment ->
            prerr_string "Comment not terminated.\n"
         | lexer__Bad_char_constant ->
            prerr_string "Ill-formed character literal.\n"
         | lexer__Unterminated_string ->
            prerr_string "String literal not terminated.\n") ;
        emit_flush () ;
        exit (-2)
    | otherexc ->
        prerr_string "Other fatal error\n";
        flush std_out ;
        printexc__f (fun () -> raise otherexc) () ;
        emit_flush () ;
        exit (-3)
   end
 with
   sys__Sys_error msg ->
      prerr_string (msg^"\n") ;
      emit_flush () ;
      exit (-4)
;;


let printHelp () =
 print_string "ml2html v0.6 (c) Francois PESSAUX 1996.\n";
 print_string "Please consult manual entry for more details on options.\n" ;
 print_string "-noanchor ; -stdout ; -version ; -noheader\n" ;
 print_string "-intcolor 24 bits hexadecimal value\n" ;
 print_string "-floatcolor 24 bits hexadecimal value\n" ;
 print_string "-stringcolor 24 bits hexadecimal value\n" ;
 print_string "-directivecolor 24 bits hexadecimal value\n" ;
 print_string "-directiveargcolor 24 bits hexadecimal value\n" ;
 print_string "-commentcolor 24 bits hexadecimal value\n" ;
 print_string "-help (or) -? (or) -options\n" ;
 exit (0)
;;


let main () =
 let rec (filestoparse:string list ref) = ref []
 and convert_filelist =
      function [] -> exit 0 | h::q -> convert_file h ; convert_filelist q
 in
  arg__parse ["-noanchor", arg__Unit (fun () -> misc__anchorflag:=false) ;
              "-stdout", arg__Unit (fun () -> misc__stdoutflag:=true) ;
              "-version", arg__Unit (fun () ->
                    print_string "ml2html v0.4 (c) Francois PESSAUX 1996.\n";
                    exit 0) ;
              "-noheader", arg__Unit (fun () -> misc__headerflag:=false) ;
              "-intcolor", arg__String (fun s ->
                      if isHexaString s then intcolor := s
                      else 
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-5)
                       end) ;
              "-floatcolor", arg__String (fun s ->
                      if isHexaString s then floatcolor := s
                      else 
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-6)
                       end) ;
              "-stringcolor", arg__String (fun s ->
                      if isHexaString s then stringcolor := s
                      else 
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-6)
                       end);
              "-directivecolor", arg__String (fun s ->
                      if isHexaString s then directivecolor := s
                      else 
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-6)
                       end);
              "-directiveargcolor", arg__String (fun s ->
                      if isHexaString s then directiveargcolor := s
                      else
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-6)
                       end);
              "-commentcolor", arg__String (fun s ->
                      if isHexaString s then commentcolor := s
                      else
                       begin
                       prerr_string ("\""^s^"\""^
                         " isn't a 24 bits hexadecimal color value.\n");
                       exit (-6)
                       end);
              "-options", arg__Unit printHelp ;
              "-help", arg__Unit printHelp ;
              "-?", arg__Unit printHelp]
             (fun filename ->filestoparse:=!filestoparse@[filename]) ;
  convert_filelist !filestoparse
;;

main() ;;

