#open "treetype" ;;

(* GLOBAL OUTPUT CHANNEL *)
let current_out_channel = ref stdout ;;
(* GLOBAL FLAGS *)
let anchorflag = ref true ;;
let stdoutflag = ref false ;;
let headerflag = ref true ;;
(* GLOBAL COLOR VALUES *)
let intcolor = ref "FF0000" ;;
let floatcolor = ref "005787" ;;
let stringcolor = ref "005000" ;;
let charcolor = ref "F2007F" ;;
let directivecolor = ref "B54A00" ;;
let directiveargcolor = ref "005000" ;;
let commentcolor = ref "505050" ;;



(* DON'T ETA-REDUCE, IT SEEMS TO CONFUSE SOMETHING *)
let emit_string s = output_string !current_out_channel s ;;
let emit_int i = output_string !current_out_channel (string_of_int i) ;;
let emit_newline () = output_string !current_out_channel "\n" ;;
let emit_float f = output_string !current_out_channel (string_of_float f) ;;
let emit_flush () = flush !current_out_channel ;;
let emit_space () = output_string !current_out_channel " " ;;
(* EMIT A SEPARATED-LIST AND KEEP TRACE OF DEFINITIONS *)
let emit_list fprint l =
 let rec recemit defs = function
    [] -> defs
  | h::q ->
     let newdefs = (begin
      match h with
         Lel_elem e ->
            fprint e
       | Lel_sep s ->
            emit_string s ;
            []
      end)
      in recemit (newdefs@defs) q
 in recemit [] l
;;

(* EMIT A SIMPLE LIST (IN CAML MEANING) AND KEEP TRACE OF DEFINITIONS *)
let emit_simplelist fprint l =
 let rec recemit defs = function
    [] -> defs
  | h::q ->
     let newdefs = fprint h
      in recemit (newdefs@defs) q
 in recemit [] l
;;

(* EMIT A SEPARATED-LIST AND DON'T BOTHER WITH DEFINITIONS *)
let do_seplist fprint l =
 let rec recemit = function
    [] -> ()
  | h::q ->
     (begin
     match h with
        Lel_elem e -> fprint e
      | Lel_sep s -> emit_string s
     end) ;
     recemit q
 in recemit l
;;

let emit_href_local anchor text =
 if !anchorflag
  then emit_string ("<A HREF=\"#"^anchor^"\">"^text^"</A>")
  else emit_string text
;;

let emit_aname anchor text =
 if !anchorflag
  then emit_string ("<A NAME=\""^anchor^"\">"^text^"</A>")
  else emit_string text
;;

type Object_type =
   Obj_value   (* VALUES AND EXCEPTIONS ARE JOINED *)
 | Obj_type
;;

let (envdef:(string*Object_type*int) list ref) = ref [] ;;

let (newmarker,resetmarker) =
 let cpt = ref 0 in
  ((function () -> cpt := !cpt + 1 ;
  !cpt),
   (function () -> cpt := 0))
;;

let is_defined str objtype env  =
 let rec recis_defined = function
    [] -> None
  | (name,objtyp,mark)::q ->
       if (name=str & objtyp=objtype)
        then Some (mark)
        else recis_defined q
 in recis_defined env
;;


let isHexaString s =
 let isHexaLetter = function
    `0` .. `9` -> true
  | `a` .. `f` -> true
  | `A` .. `F` -> true
  | _ -> false
 in
 try
  isHexaLetter (nth_char s 0) && isHexaLetter (nth_char s 1) &&
  isHexaLetter (nth_char s 2) && isHexaLetter (nth_char s 3) &&
  isHexaLetter (nth_char s 3) && isHexaLetter (nth_char s 5)
 with
  Invalid_argument "nth_char" -> false
;;
