/* The parser definition */

%{
#open "treetype";;
%}

/* Tokens */

/* Identifiers, prefixes, infixes */
%token <string*string> IDENT
%token <string*string> PREFIX
%token <string*string> INFIX0
%token <string*string> INFIX1
%token <string*string> INFIX2
%token <string*string> SUBTRACTIVE
%token <string*string> INFIX3
%token <string*string> INFIX4
/* Literals */
%token <int*string> INT
%token <string*string> CHAR
%token <float*string> FLOAT
%token <string*string> STRING
/* The end-of-file marker */
%token <string> EOF
/* Special symbols */
%token <string> EQUAL          /* "=" */
%token <string> EQUALEQUAL     /* "==" */
%token <string> SHARP          /* "#" */
%token <string> AMPERSAND      /* "&" */
%token <string> QUOTE          /* "'" */
%token <string> LPAREN         /* "(" */
%token <string> RPAREN         /* ")" */
%token <string> STAR           /* "*" */
%token <string> COMMA          /* "," */
%token <string> MINUSGREATER   /* "->" */
%token <string> DOT            /* "." */
%token <string> DOTDOT         /* ".." */
%token <string> DOTLPAREN      /* ".(" */
%token <string> DOTLBRACKET    /* ".[" */
%token <string> COLON          /* ":" */
%token <string> COLONCOLON     /* "::" */
%token <string> COLONEQUAL     /* ":=" */
%token <string> SEMI           /* ";" */
%token <string> SEMISEMI       /* ";;" */
%token <string> LBRACKET       /* "[" */
%token <string> LBRACKETBAR    /* "[|" */
%token <string> LBRACKETLESS   /* "[<" */
%token <string> LESSMINUS      /* "<-" */
%token <string> RBRACKET       /* "]" */
%token <string> UNDERSCORE     /* "_" */
%token <string> UNDERUNDER     /* "__" */
%token <string> LBRACE         /* "{" */
%token <string> BAR            /* "|" */
%token <string> BARRBRACKET    /* "|]" */
%token <string> GREATERRBRACKET/* ">]" */
%token <string> RBRACE         /* "}" */
%token <string> AMPERAMPER     /* "&&" */
%token <string> BARBAR         /* "||" */
/* Keywords */
%token <string> AND            /* "and" */
%token <string> AS             /* "as" */
%token <string> BEGIN          /* "begin" */
%token <string> DO            /* "do" */
%token <string> DONE           /* "done" */
%token <string> DOWNTO         /* "downto" */
%token <string> ELSE           /* "else" */
%token <string> END            /* "end" */
%token <string> EXCEPTION      /* "exception" */
%token <string> FOR            /* "for" */
%token <string> FUN            /* "fun" */
%token <string> FUNCTION       /* "function" */
%token <string> IF             /* "if" */
%token <string> IN             /* "in" */
%token <string> LET            /* "let" */
%token <string> MATCH          /* "match" */
%token <string> MUTABLE        /* "mutable" */
%token <string> NOT            /* "not" */
%token <string> OF             /* "of" */
%token <string> OR             /* "or" */
%token <string> PREF           /* "prefix" */
%token <string> REC            /* "rec" */
%token <string> THEN           /* "then" */
%token <string> TO             /* "to" */
%token <string> TRY            /* "try" */
%token <string> TYPE           /* "type" */
%token <string> VALUE          /* "value" */
%token <string> WHEN           /* "when" */
%token <string> WHERE          /* "where" */
%token <string> WHILE          /* "while" */
%token <string> WITH           /* "with" */

/* Precedences and associativities. Lower precedences first. */

%right prec_let
%right prec_define
%right MINUSGREATER
%right WHERE
%right AND
%right SEMI
%right prec_list
%right prec_if
%right COLONEQUAL LESSMINUS
%left  AS
%left  BAR
%left  COMMA
%left  OR BARBAR
%left  AMPERSAND AMPERAMPER
%left  NOT
%left  INFIX0 EQUAL EQUALEQUAL          /* comparisons */
%right INFIX1                           /* concatenations */
%right COLONCOLON                       /* cons */
%left  INFIX2 SUBTRACTIVE               /* additives, subtractives */
%left  STAR INFIX3                      /* multiplicatives */
%right INFIX4                           /* exponentiations */
%right prec_uminus
%right prec_app
%left  DOT DOTLPAREN DOTLBRACKET
%right PREFIX                           /* prefix operators, e.g. ! */

/* Entry points */

%start Implementation
%type <treetype__At_implent> Implementation 

%%

/* One phrase from a module implementation */

Implementation :
        Expr SEMISEMI
          { Imp_expr ($1,$2) }
      | LET Binding_list SEMISEMI  %prec prec_let
          { Imp_let ($2,$1,$3) }
      | LET REC Binding_list SEMISEMI  %prec prec_let
          { Imp_letrec ($3,$1,$2,$4) }
      | TYPE Type_decl SEMISEMI
          { Imp_type ($2,$1,$3) }
      | EXCEPTION Exc_decl SEMISEMI
          { Imp_exception ($2,$1,$3) }
      | SHARP Directive SEMISEMI
          { Imp_directive ($2,$1,$3) }
      | EOF
          { raise End_of_file }
;

/* One phrase from a module interface */


/* Expressions */

Expr :
        Simple_expr
          { Expr_sexpr $1 }
      | Simple_expr Simple_expr_list   %prec prec_app
          { Expr_sexprlst ($1::$2) }
      | Expr_comma_list
          { Expr_exprcomma $1 }
      | SUBTRACTIVE Expr  %prec prec_uminus
          { Expr_substr1 ((fst $1),$2,(snd $1)) }
      | NOT Expr { Expr_not ($2,$1) }
      | Ide LESSMINUS Expr { Expr_lessminus ($1,$3,$2) }
      | Expr INFIX4 Expr { Expr_infix ((fst $2),$1,$3,(snd $2)) }
      | Expr INFIX3 Expr { Expr_infix ((fst $2),$1,$3,(snd $2)) }
      | Expr INFIX2 Expr { Expr_infix ((fst $2),$1,$3,(snd $2)) }
      | Expr SUBTRACTIVE Expr { Expr_substr2 ((fst $2),$1,$3,(snd $2)) }
      | Expr INFIX1 Expr { Expr_infix ((fst $2),$1,$3,(snd $2)) }
      | Expr INFIX0 Expr { Expr_infix ((fst $2),$1,$3,(snd $2)) }
      | Expr STAR Expr { Expr_star ($1,$3,$2) }
      | Expr COLONCOLON Expr { Expr_coloncolon ($1,$3,$2) }
      | Expr EQUAL Expr { Expr_equal ($1,$3,$2) }
      | Expr EQUALEQUAL Expr { Expr_equalequal ($1,$3,$2) }
      | Expr AMPERSAND Expr { Expr_ampersand ($1,$3,$2) }
      | Expr AMPERAMPER Expr { Expr_amperamper ($1,$3,$2) }
      | Expr OR Expr { Expr_or ($1,$3,$2) }
      | Expr BARBAR Expr { Expr_barbar ($1,$3,$2) }
      | Simple_expr DOT Ext_ident LESSMINUS Expr
          { Expr_dotlessminus ($1,$3,$5,$2,$4) }
      | Simple_expr DOTLPAREN Expr RPAREN LESSMINUS Expr
          { Expr_dotlparen ($1,$3,$6,$2,$4,$5) }
      | Simple_expr DOTLBRACKET Expr RBRACKET LESSMINUS Expr
          { Expr_dotlbracket ($1,$3,$6,$2,$4,$5) }
      | Expr COLONEQUAL Expr
          { Expr_colonequal ($1,$3,$2) }
      | IF Expr THEN Expr ELSE Expr  %prec prec_if
          { Expr_ifthenelse ($2,$4,$6,$1,$3,$5) }
      | IF Expr THEN Expr  %prec prec_if
          { Expr_ifthen ($2,$4,$1,$3) }
      | WHILE Expr DO Opt_expr DONE
          { Expr_while ($2,$4,$1,$3,$5) }
      | FOR Ide EQUAL Expr TO Expr DO Opt_expr DONE
          { Expr_forup ($2,$4,$6,$8,$1,$3,$5,$7,$9) }
      | FOR Ide EQUAL Expr DOWNTO Expr DO Opt_expr DONE
          { Expr_fordown ($2,$4,$6,$8,$1,$3,$5,$7,$9) }
      | Expr SEMI Expr
          { Expr_semisemi ($1,$3,$2) }
      | Expr SEMI
          { Expr_semi ($1,$2) }
      | MATCH Expr WITH Opt_bar Function_match
          { Expr_match ($2,$5,$1,$3,$4) }
      | MATCH Expr WITH Opt_bar Parser_match
          { Expr_matchstream ($2,$5,$1,$3,$4) }
      | LET Binding_list IN Expr  %prec prec_let
          { Expr_let ($2,$4,$1,$3) }
      | LET REC Binding_list IN Expr  %prec prec_let
          { Expr_letrec ($3,$5,$1,$2,$4) }
      | FUN Opt_bar Fun_match
          { Expr_fun ($3,$1,$2) }
      | FUNCTION Opt_bar Function_match
          { Expr_function ($3,$1,$2) }
      | FUNCTION Opt_bar Parser_match
          { Expr_functionstream ($3,$1,$2) }
      | TRY Expr WITH Opt_bar Try_match
          { Expr_try ($2,$5,$1,$3,$4) }
      | Expr WHERE Binding_list
          { Expr_where ($1,$3,$2) }
      | Expr WHERE REC Binding_list  %prec WHERE
          { Expr_whererec ($1,$4,$2,$3) }
;

Simple_expr :
        Struct_constant
          { Sexp_structcst $1 }
      | Ext_ident
          { Sexp_extident $1 }
      | LBRACKET Expr_sm_list RBRACKET
          { Sexp_list ($2,$1,$3) }
      | LBRACKETBAR Expr_sm_list BARRBRACKET
          { Sexp_vector ($2,$1,$3) }
      | LBRACKETLESS Stream_expr GREATERRBRACKET
          { Sexp_stream ($2,$1,$3) }
      | LPAREN Expr COLON Type RPAREN
          { Sexp_withtype ($2,$4,$1,$3,$5) }
      | LPAREN Opt_expr RPAREN
          { Sexp_paren ($2,$1,$3) }
      | BEGIN Opt_expr END
          { Sexp_begin ($2,$1,$3) }
      | LBRACE Expr_label_list RBRACE
          { Sexp_record ($2,$1,$3) }
      | PREFIX Simple_expr
          { Sexp_prefix ((fst $1),$2,(snd $1)) }
      | Simple_expr DOT Ext_ident
          { Sexp_dot ($1,$3,$2) }
      | Simple_expr DOTLPAREN Expr RPAREN
          { Sexp_dotlparen ($1,$3,$2,$4) }
      | Simple_expr DOTLBRACKET Expr RBRACKET
          { Sexp_dotlbracket ($1,$3,$2,$4) }
;

Simple_expr_list :
        Simple_expr Simple_expr_list
          { $1::$2 }
      | Simple_expr
          { [$1] }
;

Expr_comma_list :
        Expr_comma_list COMMA Expr
          { $1@[(Lel_sep ($2^"<B>,</B>"));(Lel_elem $3)] }
      | Expr COMMA Expr
          { [(Lel_elem $1);(Lel_sep ($2^"<B>,</B>"));(Lel_elem $3)] }
;

Expr_sm_list :
        Expr_sm_list SEMI Expr  %prec prec_list
          { $1@[(Lel_sep ($2^"<B>;</B>"));(Lel_elem $3)] }
      | Expr                    %prec prec_list
          { [(Lel_elem $1)] }
      | Expr_sm_list SEMI
          { $1@[(Lel_sep ($2^"<B>;</B>"))] }
      | /*epsilon*/
          { [] }
;

Opt_expr :
        Expr
          { Some $1 }
      | /*epsilon*/
          { None }

Expr_label :
        Ext_ident EQUAL Expr
          { ($1,$3,$2) }
;

Expr_label_list :
        Expr_label SEMI Expr_label_list
          { (Lel_elem $1)::(Lel_sep ($2^"<B>;</B>"))::$3 }
      | Expr_label
          { [(Lel_elem $1)] }
      | Expr_label SEMI
          { [(Lel_elem $1);(Lel_sep ($2^"<B>;</B>"))] }
;

/* Constants */
Struct_constant :
        Atomic_constant { $1 }
;

Atomic_constant :
        INT { Acst_int $1 }
      | FLOAT { Acst_float $1 }
      | STRING { Acst_string $1 }
      | CHAR { Acst_char $1 }
;

/* Definitions by pattern matchings */

Opt_bar :
        BAR { $1^"<B>|</B>" }
      | /*epsilon*/ { "" }
;

Action :
        MINUSGREATER Expr
          { Act_direct ($2,$1) }
      | WHEN Expr MINUSGREATER Expr
          { Act_when ($2,$4,$1,$3) }
;

Fun_match :
        Simple_pattern_list Action BAR Fun_match
          { [Lel_elem ($1,$2);(Lel_sep ($3^"<B>|</B>"))]@$4 }
      | Simple_pattern_list Action
          { [Lel_elem ($1,$2)] }
;

Function_match :
        Pattern Action BAR Function_match
          { [Lel_elem ($1,$2);(Lel_sep ($3^"<B>|</B>"))]@$4 }
      | Pattern Action
          { [Lel_elem ($1,$2)] }
;

Try_match :
        Pattern Action BAR Try_match
          { (Lel_elem ($1,$2))::(Lel_sep ($3^"<B>|</B>"))::$4 }
      | Pattern Action
          { [Lel_elem ($1,$2)] }
;

Binding_list :
        Binding AND Binding_list
          { (Lel_elem $1)::(Lel_sep ($2^"<B>and</B>"))::$3 }
      | Binding
          { [(Lel_elem $1)] }
;

Binding :
        Pattern EQUAL Expr  %prec prec_define
          { Bind_pattern ($1,$3,$2) }
      | Ide Simple_pattern_list EQUAL Expr  %prec prec_define
          { Bind_ident ($1,$2,$4,$3) }
;

/* Patterns */
Pattern_sm_list :
        Pattern_sm_list SEMI Pattern
          { $1@[(Lel_sep ($2^"<B>;</B>"));(Lel_elem $3)] }
      | Pattern
          { [Lel_elem $1] }
      | Pattern_sm_list SEMI
          { $1@[Lel_sep ($2^"<B>;</B>")] }
      | /*epsilon*/
          { [] }
;

Pattern_label_list :
        Pattern_label SEMI Pattern_label_list
          { (Lel_elem $1)::(Lel_sep ($2^"<B>;</B>"))::$3 }
      | Pattern_label
          { [(Lel_elem $1)] }
      | UNDERSCORE
          { [Lel_elem (Patl_underscore $1)] }
      | /*epsilon*/
          { [] }
;

Pattern_label :
        Ext_ident EQUAL Pattern
          { Patl_other ($1,$3,$2) }
;

Pattern_comma_list :
        Pattern_comma_list COMMA Pattern
          { $1@[(Lel_sep ($2^"<B>,</B>"));(Lel_elem $3)] }
      | Pattern COMMA Pattern
          { [(Lel_elem $1);(Lel_sep ($2^"<B>,</B>"));(Lel_elem $3)] }
;
  
Simple_pattern_list :
        Simple_pattern Simple_pattern_list
          { $1::$2 }
      | Simple_pattern
          { [$1] }
;

Pattern :
        Simple_pattern
          { Pat_spat $1 }
      | Pattern AS IDENT
          { Pat_asid ($1,(fst $3),$2,(snd $3)) }
      | Pattern COLONCOLON Pattern
          { Pat_coloncolon ($1,$3,$2) }
      | Pattern_comma_list
          { Pat_commalst $1 }
      | Ext_ident Simple_pattern
          { Pat_extid ($1,$2) }
      | Pattern BAR Pattern
          { Pat_multi ($1,$3,$2) }
;

Simple_pattern :
        Atomic_constant
          { Spat_atcst $1 }
      | SUBTRACTIVE INT
          { Spat_subsint ((fst $1),(fst $2),(snd $1),(snd $2)) }
      | SUBTRACTIVE FLOAT
          { Spat_subfloat ((fst $1),(fst $2),(snd $1),(snd $2)) }
      | UNDERSCORE
          { Spat_underscore $1 }
      | Ide
          { Spat_ident $1 }
      | Qual_ident
          { Spat_qualid $1 }
      | LPAREN RPAREN
          { Spat_unit ($1,$2) }
      | LBRACKET Pattern_sm_list RBRACKET
          { Spat_list ($2,$1,$3)}
      | LPAREN Pattern COLON Type RPAREN
          { Spat_withtype ($2,$4,$1,$3,$5) }
      | LBRACE Pattern_label_list RBRACE
          { Spat_record ($2,$1,$3) }
      | LPAREN Pattern RPAREN
          { Spat_paren ($2,$1,$3) }
      | CHAR DOTDOT CHAR
          { Spat_charinterv ((fst $1),(fst $3),$2,(snd $1),(snd $3)) }
;

/* Streams */
Stream_expr :
        Stream_expr SEMI Stream_expr_component
          { $1@[(Lel_sep ($2^"<B>;</B>"));(Lel_elem $3)] }
      | Stream_expr_component
          { [Lel_elem $1] }
      | Stream_expr SEMI
          { $1@[(Lel_sep ($2^"<B>;</B>"))] }
      | /*epsilon*/
          { [] }
;

Stream_expr_component :
        Expr %prec prec_list
          { Sec_expr $1 }
      | QUOTE Expr  %prec prec_list
          { Sec_quote ($2,$1) }
;

Stream_pattern :
        LBRACKETLESS Stream_pattern_component_list GREATERRBRACKET
          { Strp_full ($2,$1,$3) }
      | LBRACKETLESS GREATERRBRACKET
          { Strp_empty ($1,$2) }
;

Stream_pattern_component_list :
        IDENT
          { [Lel_elem (Spc_ident $1)] }
      | Stream_pattern_component SEMI Stream_pattern_component_list
          { (Lel_elem $1)::(Lel_sep ($2^"<B>;</B>"))::$3 }
      | Stream_pattern_component
          { [Lel_elem $1] }
;

Stream_pattern_component :
        Simple_expr Simple_pattern
          { Spc_sexprspat ($1,$2) }
      | QUOTE Pattern
          { Spc_quote ($2,$1) }
      | Stream_pattern_component SEMI
          { Spc_semi ($1,$2) }
;

Parser_match :
        Stream_pattern MINUSGREATER Expr BAR Parser_match
          { [Lel_elem ($1,$3,$2);(Lel_sep ($4^"<B>|</B>"))]@$5 }
      | Stream_pattern MINUSGREATER Expr
          { [Lel_elem ($1,$3,$2)] }
;

/* Identifiers */
Ide :
        IDENT { Idnt_simple $1 }
      | PREF Infx { Idnt_prefix ((fst $2),$1,(snd $2)) }
;

Infx :
        INFIX0          { $1 }
      | INFIX1          { $1 }
      | INFIX2          { $1 }
      | INFIX3          { $1 }
      | INFIX4          { $1 }
      | STAR            { ("*",$1) }
      | COLONCOLON      { ("::",$1) }
      | COLONEQUAL      { (":=",$1) }
      | EQUAL           { ("=",$1) }
      | EQUALEQUAL      { ("==",$1) }
      | NOT             { ("not",$1) }
      | SUBTRACTIVE     { $1 }
      | PREFIX          { $1 }
      | AMPERSAND       { ("&amp;",$1) }
      | AMPERAMPER      { ("&amp;&amp;",$1) }
      | OR              { ("or",$1) }
      | BARBAR          { ("||",$1) }
;

Qual_ident :
        IDENT UNDERUNDER Ide
          { ((fst $1),$3,(snd $1),$2) }
;

Ext_ident :
        Qual_ident
          { Eid_qual $1 }
      | Ide
          { Eid_simple $1 }
;

/* Type expressions */
Type :
        Simple_type
          { Typ_simple $1 }
      | Type_star_list
          { Typ_starlst $1 }
      | Type MINUSGREATER Type
          { Typ_minusgreater ($1,$3,$2) }
;

Simple_type :
        Type_var
          { Styp_var $1 }
      | Ext_ident
          { Styp_extid $1 }
      | Simple_type Ext_ident
          { Styp_typapp ($1,$2) }
      | LPAREN Type COMMA Type_comma_list RPAREN Ext_ident
          { Styp_typappmulti ($2,$4,$6,$1,$3,$5) }
      | LPAREN Type RPAREN
          { Styp_paren ($2,$1,$3) }
;

Type_star_list :
        Type_star_list STAR Simple_type
          { $1@[(Lel_sep ($2^"<B>*</B>"));(Lel_elem $3)] }
      | Simple_type STAR Simple_type
          { [(Lel_elem $1);(Lel_sep ($2^"<B>*</B>"));(Lel_elem $3)] }
;

Type_var :
        QUOTE IDENT
          { ((fst $2),($1),(snd $2)) }
;

Type_comma_list :
        Type COMMA Type_comma_list
          { (Lel_elem $1)::(Lel_sep $2)::$3 }
      | Type
          { [(Lel_elem $1)] }
;

/* Declarations */

Value_decl :
        Value1_decl AND Value_decl
          { $1::$3 }
      | Value1_decl
          { [$1] }
;

Type_decl :
        Type1_decl AND Type_decl
          { (Lel_elem $1)::(Lel_sep ($2^"<B>and</B>"))::$3 }
      | Type1_decl
          { [Lel_elem $1] }
;

Exc_decl :
        Constr1_decl AND Exc_decl
          { (Lel_elem $1)::(Lel_sep ($2^"<B>and</B>"))::$3 }
      | Constr1_decl
          { [Lel_elem $1] }
;

Constr_decl :
        Constr1_decl BAR Constr_decl
          { (Lel_elem $1)::(Lel_sep ($2^"<B>|</B>"))::$3 }
      | Constr1_decl
          { [Lel_elem $1] }
;

Label_decl :
        Label1_decl SEMI Label_decl
          { (Lel_elem $1)::(Lel_sep ($2^"<B>;</B>"))::$3 }
      | Label1_decl SEMI
          { [(Lel_elem $1);(Lel_sep ($2^"<B>;</B>"))] }
      | Label1_decl
          { [Lel_elem $1] }
;

Value1_decl :
        Ide COLON Type
           { Vdecl_simple ($1,$3) }
      | Ide COLON Type EQUAL Prim_decl
           { Vdecl_equal ($1,$3,$5) }
;

Prim_decl :
        INT STRING
           { ((fst $1),(fst $2)) }
;

Type1_decl :
        Type_params IDENT Type1_def
          { ($1,(fst $2),$3,(snd $2)) }
;

Type1_def :
        /* epsilon */
          { Tdef_abstract }
      | EQUAL Opt_bar Constr_decl
          { Tdef_sum ($3,$1,$2) }
      | EQUAL LBRACE Label_decl RBRACE
          { Tdef_record ($3,$1,$2,$4) }
      | EQUALEQUAL Type
          { Tdef_abrev ($2,$1) }
;

Constr1_decl :
        IDENT OF Mutable_option Type
          { Tc_arityn ((fst $1),$3,$4,(snd $1),$2) }
      | IDENT
          { Tc_arity0 $1 }
;

Label1_decl :
        Mutable_option IDENT COLON Type
          { ($1,(fst $2),$4,(snd $2),$3) }
;

Mutable_option :
        MUTABLE
          { Mut_yes $1 }
      | /* epsilon */
          { Mut_no }
;

Type_params :
        LPAREN Type_var_list RPAREN
          { Tpar_paren ($2,$1,$3) }
      | Type_var
          { Tpar_var $1 }
      | /* epsilon */
          { Tpar_none }
;

Type_var_list :
        Type_var COMMA Type_var_list
          { (Lel_elem $1)::(Lel_sep ($2^"<B>,</B>"))::$3 }
      | Type_var
          { [Lel_elem $1] }
;
       
/* Directives */
Directive :
        IDENT STRING
          { ($1,$2) }
;

%%

