(* Last string arguments for constructors are text separation BEFORE *)
(* each part of the token *)

exception End_of_file ;;

type 'a option = None | Some of 'a ;;

(* Type for list elemts separated by separator text *)
type 'a Li_listelem =
   Lel_sep of string
 | Lel_elem of 'a
;;

(* At : Abstract Tree *)
type At_atomcst =
   (* Acst : Atomic Constant *)
   Acst_int of (int*string)
 | Acst_float of (float*string)
 | Acst_string of (string*string)
 | Acst_char of (string*string)  (* A CHAR IS VIEWED LIKE A STRING *)

and At_ident =
   Idnt_simple of (string*string)
 | Idnt_prefix of (string*string*string)

and  At_extid =
   Eid_qual of (string*At_ident*string*string)
 | Eid_simple of At_ident

and  At_stype =
   (* Styp : Simple Type *)
   Styp_var of (string*string*string)
 | Styp_extid of At_extid
 | Styp_typapp of (At_stype*At_extid)
 | Styp_typappmulti of (At_type*((At_type Li_listelem) list)*
                       At_extid*string*string*string)
 | Styp_paren of (At_type*string*string)

and  At_type =
   (* Typ : Type *)
   Typ_simple of At_stype
 | Typ_starlst of (At_stype Li_listelem) list
 | Typ_minusgreater of (At_type*At_type*string)

and  At_patlabel =
   Patl_underscore of string
 | Patl_other of (At_extid*At_pattern*string)

and  At_spattern =
   (* Spat : Simple Pattern *)
   Spat_atcst of At_atomcst
 | Spat_subsint of (string*int*string*string)
 | Spat_subfloat of (string*float*string*string)
 | Spat_underscore of string
 | Spat_ident of At_ident
 | Spat_qualid of (string*At_ident*string*string)
 | Spat_unit of (string*string)
 | Spat_list of ((At_pattern Li_listelem) list)*string*string
 | Spat_withtype of (At_pattern*At_type*string*string*string)
 | Spat_record of ((At_patlabel Li_listelem) list)*string*string
 | Spat_paren of (At_pattern*string*string)
 | Spat_charinterv of (string*string*string*string*string)

and  At_pattern =
   (* Pat : Pattern *)
   Pat_spat of At_spattern
 | Pat_asid of (At_pattern*string*string*string)
 | Pat_coloncolon of (At_pattern*At_pattern*string)
 | Pat_commalst of (At_pattern Li_listelem) list
 | Pat_extid of (At_extid*At_spattern)
 | Pat_multi of (At_pattern*At_pattern*string)

and  At_expr =
   (* Expr : Expression *)
   Expr_sexpr of At_sexpr
 | Expr_sexprlst of At_sexpr list
 | Expr_exprcomma of (At_expr Li_listelem) list
 | Expr_substr1 of (string*At_expr*string)
 | Expr_not of (At_expr*string)
 | Expr_lessminus of (At_ident*At_expr*string)
 | Expr_infix of (string*At_expr*At_expr*string)
 | Expr_substr2 of (string*At_expr*At_expr*string)
 | Expr_star of (At_expr*At_expr*string)
 | Expr_coloncolon of (At_expr*At_expr*string)
 | Expr_equal of (At_expr*At_expr*string)
 | Expr_equalequal of (At_expr*At_expr*string)
 | Expr_ampersand of (At_expr*At_expr*string)
 | Expr_amperamper of (At_expr*At_expr*string)
 | Expr_or of (At_expr*At_expr*string)
 | Expr_barbar of (At_expr*At_expr*string)
 | Expr_dotlessminus of (At_sexpr*At_extid*At_expr*string*string)
 | Expr_dotlparen of (At_sexpr*At_expr*At_expr*string*string*string)
 | Expr_dotlbracket of (At_sexpr*At_expr*At_expr*string*string*string)
 | Expr_colonequal of (At_expr*At_expr*string)
 | Expr_ifthenelse of (At_expr*At_expr*At_expr*string*string*string)
 | Expr_ifthen of (At_expr*At_expr*string*string)
 | Expr_while of (At_expr*(At_expr option)*string*string*string)
 | Expr_forup of (At_ident*At_expr*At_expr*(At_expr option)*
          string*string*string*string*string)
 | Expr_fordown of (At_ident*At_expr*At_expr*(At_expr option)*
          string*string*string*string*string)
 | Expr_semisemi of (At_expr*At_expr*string)
 | Expr_semi of (At_expr*string)
 | Expr_match of (At_expr*(((At_pattern*At_action) Li_listelem) list)*
          string*string*string)
 | Expr_matchstream of (At_expr*(((At_streampat*At_expr*string)
          Li_listelem) list)*string*string*string)
 | Expr_let of (((At_binding Li_listelem) list)*At_expr*string*string)
 | Expr_letrec of (((At_binding Li_listelem) list)*At_expr*
                  string*string*string)
 | Expr_fun of ((((At_spattern list)*At_action) Li_listelem) list)*
               string*string
 | Expr_function of ((((At_pattern*At_action) Li_listelem) list)*
               string*string)
 | Expr_functionstream of ((((At_streampat*At_expr*string) Li_listelem) list)*
             string*string)
 | Expr_try of (At_expr*(((At_pattern*At_action) Li_listelem) list)*
               string*string*string)
 | Expr_where of (At_expr*((At_binding Li_listelem) list)*string)
 | Expr_whererec of (At_expr*((At_binding Li_listelem) list)*string*string)

and  At_sexpr =
   (* Sexp : Simple Expression *)
   Sexp_structcst of At_atomcst
 | Sexp_extident of At_extid
 | Sexp_list of (((At_expr Li_listelem) list)*string*string)
 | Sexp_vector of (((At_expr Li_listelem) list)*string*string)
 | Sexp_stream of (((At_streamexprcomp Li_listelem) list)*string*string)
 | Sexp_withtype of (At_expr*At_type*string*string*string)
 | Sexp_paren of ((At_expr option)*string*string)
 | Sexp_begin of ((At_expr option)*string*string)
 | Sexp_record of ((((At_extid*At_expr*string)Li_listelem)list)*string*string)
 | Sexp_prefix of (string*At_sexpr*string)
 | Sexp_dot of (At_sexpr*At_extid*string)
 | Sexp_dotlparen of (At_sexpr*At_expr*string*string)
 | Sexp_dotlbracket of (At_sexpr*At_expr*string*string)

and  At_action =
   (* Act : Action *)
   Act_direct of (At_expr*string)
 | Act_when of (At_expr*At_expr*string*string)

and  At_streamexprcomp =
   (* Sec : Stream Expression Component *)
   Sec_expr of At_expr
 | Sec_quote of (At_expr*string)

and  At_streampatcomp =
   (* Sec : Stream Pattern Component *)
   Spc_sexprspat of (At_sexpr*At_spattern)
 | Spc_quote of (At_pattern*string)
   (* Belongs to : Stream_pattern_component_list *)
 | Spc_ident of (string*string)
 | Spc_semi of (At_streampatcomp*string)

and  At_streampat =
   (* Strp : Stream Pattern *)
   Strp_full of ((At_streampatcomp Li_listelem) list)*string*string
 | Strp_empty of (string*string)

and  At_binding =
   (* Bind = Binding *)
   Bind_pattern of (At_pattern*At_expr*string)
 | Bind_ident of (At_ident*(At_spattern list)*At_expr*string)

and  At_typepar =
   Tpar_paren of ((((string*string*string) Li_listelem) list)*string*string)
 | Tpar_var of (string*string*string)
 | Tpar_none

and  At_implent =
   (* Imp : Implementation *)
   Imp_expr of (At_expr*string)
 | Imp_let of (((At_binding Li_listelem) list)*string*string)
 | Imp_letrec of (((At_binding Li_listelem) list)*string*string*string)
 | Imp_type of ((((At_typepar*string*At_typedef*string) Li_listelem) list)*
               string*string)
 | Imp_exception of (((At_constr Li_listelem) list)*string*string)
 | Imp_directive of (((string*string)*(string*string))*string*string)

and  At_mutable =
   (* Mut : Mutable *)
   Mut_yes of string
 | Mut_no

and  At_constr =
    (* Tc : Type Constructor *)
    Tc_arityn of (string*At_mutable*At_type*string*string)
  | Tc_arity0 of (string*string)

and  At_typedef =
   (* Tdef : Type Definition *)
   Tdef_abstract
 | Tdef_sum of (((At_constr Li_listelem) list)*string*string)
 | Tdef_record of (((At_mutable*string*At_type*string*string) Li_listelem)
                  list)*string*string*string
 | Tdef_abrev of (At_type*string)

and  At_valdecl =
   (* Vdecl : Value Declaration *)
   Vdecl_simple of (At_ident*At_type)
 | Vdecl_equal of (At_ident*At_type*(int*string))
;;
